import multiprocessing as mp

#Future: Use GLFW!
import pygame
from pygame.locals import *

import numpy as np

MOD_OPENGL = True
try :
	from OpenGL.GL import *
	from OpenGL.GL.shaders import compileShader,ShaderProgram
	from OpenGL.GLU import *
	from OpenGL.arrays import vbo #This is a class that makes it easy to use Vertex Buffer Objects.
	from OpenGL.GL.framebufferobjects import *
	from OpenGL.GL.EXT.framebuffer_object import *
	#~ from OpenGLContext.arrays import *
except :
	print('Unable to load OpenGL. Make sure your graphics drivers are installed & up to date!')
	MOD_OPENGL = False
	

import sys, os, os.path

class Viewer :
	def __init__(self, res, title="OpenLUT Image Viewer") :
		self.res = res
		
		#Vertex shaders calculate vertex positions - gl_position, which is a vec4.
		#In our case, this vec4 is on a ortho projected square in front of the screen.
		#~ self.shaderVertex = compileShader("""#version 330 core
#~ layout (location = 0) in vec2 position;
#~ layout (location = 1) in vec2 texCoords;

#~ out vec2 TexCoords;

#~ void main()
#~ {
    #~ gl_Position = vec4(position.x, position.y, 0.0f, 1.0f); 
    #~ TexCoords = texCoords;
#~ }  
#~ """, GL_VERTEX_SHADER )
		
		#After a vertex is processed, clupping happens, etc. Then frag shader.
		#Fragment shaders make "fragments" - pixels, subpixels, hidden stuff, etc. . They can do per pixel stuff.
		#Goal: Make gl_FragColor, the color of the fragment. It's a vec4.
		#In this case, we're sampling the texture coordinates.
		#~ self.shaderFrag = compileShader("""#version 330 core
#~ in vec2 TexCoords;
#~ out vec4 color;

#~ uniform sampler2D screenTexture;

#~ void main()
#~ { 
    #~ color = texture(screenTexture, TexCoords);
#~ }
#~ """, GL_FRAGMENT_SHADER )

		#Convenience for glCreateProgram, then attaches each shader via pointer, links with glLinkProgram,
		#validates with glValidateProgram and glGetProgramiv, then cleanup & return shader program.
		#~ self.shader = Viewer.shaderProgramCompile(self.shaderVertex, self.shaderFrag)
		#~ self.vbo = self.bindVBO()
				
		#Init pygame in OpenGL double-buffered mode.
		pygame.init()
		pygame.display.set_caption(title)
		pygame.display.set_mode((res), DOUBLEBUF|OPENGL)
		
		#Initialize OpenGL.
		self.initGL()
		
	def shaderProgramCompile(*shaders) :
		prog = glCreateProgram()
		for shader in shaders :
			glAttachShader(prog, shader)
		prog = ShaderProgram(prog)
		glLinkProgram(prog)
		return prog
		
	def initGL(self) :
		'''
		Initialize OpenGL.
		'''
		#Start up OpenGL in Ortho projection mode.
		glEnable(GL_TEXTURE_2D)
		
		glViewport(0, 0, self.res[0], self.res[1])
		
		glMatrixMode(GL_PROJECTION)
		glLoadIdentity()
		glOrtho(0, self.res[0], self.res[1], 0, 0, 100)
		
		glMatrixMode(GL_MODELVIEW)
		#~ glUseProgram(self.shader)
		
	def resizeWindow(self, newRes) :
		#~ print(newRes)
		self.res = newRes
		pygame.display.set_mode(self.res, DOUBLEBUF|OPENGL)
		glViewport(0, 0, self.res[0], self.res[1]) #Reset viewport
		
		glMatrixMode(GL_PROJECTION) #Modify projection matrix
		glLoadIdentity() #Load in identity matrix
		glOrtho(0, self.res[0], self.res[1], 0, 0, 100) #New projection matrix
		
		glMatrixMode(GL_MODELVIEW) #Switch back to model matrix.
		glLoadIdentity() #Load an identity matrix into the model-view matrix
		
		#~ pygame.display.flip()
				
	def drawImage(self) :
		'''
		Draws an image to the screen.
		'''
		#~ print("\r", self.res, end="", sep="")
		glBegin(GL_QUADS)
		
		glTexCoord2i(0, 0)
		glVertex2i(0, 0)
		
		glTexCoord2i(0, 1)
		glVertex2i(0, self.res[1])
		
		glTexCoord2i(1, 1)
		glVertex2i(self.res[0], self.res[1])
		
		glTexCoord2i(1, 0)
		glVertex2i(self.res[0], 0)
		
		glEnd()
		
	def bindVBO(self, verts=np.array([[0,1,0],[-1,-1,0],[1,-1,0]], dtype='f')) :
		vertPos = vbo.VBO(verts)
		
		indices = np.array([[0, 1, 2]], dtype=np.int32)
		indPos = vbo.VBO(indices, target=GL_ELEMENT_ARRAY_BUFFER)
		
		return (vertPos, indPos)
		
	def bindFBO(self) :
		'''
		Create and bind a framebuffer for rendering (loading images) to.
		'''
		
		fbo = glGenFramebuffers(1) #Create framebuffer
		
		#Binding it makes the next read and write framebuffer ops affect the bound framebuffer.
		#You can also bind it specifically to read/write targets. GL_READ_FRAMEBUFFER and GL_DRAW_FRAMEBUFFER.
		glBindFramebuffer(GL_FRAMEBUFFER, fbo)
		
		#It needs 1+ same sampled buffers (color, depth, stencil) and a "complete" color attachment.
		
		#Create a texture to render to. Empty for now; size is screen size.
		tex = self.bindTex(None, res=self.res) #Fill it up with nothing, for now. It's our color attachment.
		
		glBindTexture(GL_TEXTURE_2D, 0)
		
		#Target is framebuffer, attachment is color, textarget is 2D texture, the texture is tex, the mipmap level is 0.
		#We attach the texture to the frame buffer.
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT1, GL_TEXTURE_2D, tex, 0)
		
		#Renderbuffers are write-only; can't be sampled, just displayed. Often used as depth and stencil. So useless here :).
		
		if glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE :
			print("Framebuffer not complete!")
			
		glBindFramebuffer(GL_FRAMEBUFFER, 0); #Finally - bind the framebuffer!
		
		#We're now rendering to the framebuffer texture. How cool!
		
		return fbo
		
				
	def bindTex(self, img, res=None)	:
		'''
		Binds the image contained the numpy float array img to a 2D texture on the GPU.
		'''
		if not res: res = img.shape
		
		tex = glGenTextures(1)
		
		glPixelStorei(GL_UNPACK_ALIGNMENT, 1)
		glBindTexture(GL_TEXTURE_2D, tex)
		
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP) #Clamp to edge
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP)
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR) #Mag/Min Interpolation
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
		
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, res[1], res[0], 0, GL_RGB, GL_FLOAT, img)
		
		return tex
		
	def display(self, fbo = 1, tex = 1) :
		'''
		Repaints the window.
		'''
		
		#Here, we do things to the framebuffer. Not the screen. Important.
		#~ glBindFramebuffer(GL_FRAMEBUFFER, fbo)
		#~ glClearColor(0, 0, 0, 1.0)
		glClear(GL_COLOR_BUFFER_BIT)
		glMatrixMode(GL_MODELVIEW)
		
		#This render is rendering to framebuffer
		glEnable(GL_TEXTURE_2D)
		self.drawImage()
		
		#~ #Back to the screen.
		#~ glBindFramebuffer(GL_FRAMEBUFFER, 0)
		#~ glClearColor(0, 1, 1, 1)
		#~ glClear(GL_COLOR_BUFFER_BIT)
		
		#~ glUseProgram(self.shader)
		#~ glBindTexture(GL_TEXTURE_2D, tex)
		
		#~ glBindVertexArray(0)
		#~ glUseProgram(0)
		
		#Updates the display.
		pygame.display.flip()
		
	def close() :
		print()
		#~ glUseProgram(0)
		pygame.quit()
		
	def run(img, xRes, yRes, title = "OpenLUT Image Viewer") :
		'''
		img is an rgb array.
		'''
		if not MOD_OPENGL: print("OpenGL not enabled. Viewer won't start."); return
		
		v = Viewer((xRes, yRes), title)
		gpuImg = v.bindTex(img)
		#~ gpuBuf = v.bindFBO()
		
		FPS = None
		clock = pygame.time.Clock()
		
		while True :			
			for event in pygame.event.get() :
				if event.type == pygame.QUIT: Viewer.close(); break
				
				if event.type == pygame.VIDEORESIZE :
					v.resizeWindow((event.w, event.h))
					
				if event.type == pygame.KEYDOWN :
					if str(event.key) == "27": Viewer.close(); break #Need to catch ESC to close the window.
					
					try :
						{
							
						}[event.key]()
					except KeyError as key :
						print("Key not mapped!")
			else :
				#This else will only run if the event loop is completed.
				#~ v.display(fbo = gpuBuf, tex = gpuImg)
				v.display()
				
				#Smooth playback at FPS.
				if FPS: clock.tick(FPS)
				else: clock.tick()
				#~ print("\r", clock.get_fps(), end="", sep="")
				
				continue
				
			break #This break will only run if the event loop is broken out of.
			



import multiprocessing as mp
from functools import reduce
import operator as oper

import numpy as np
#~ import numba

from .Transform import Transform
from .lib import olOpt as olo

class ColMat(Transform) :
	def __init__(self, *mats) :
		'''
		Initializes a combined 3x3 Transformation Matrix from any number of input matrices. These may be numpy arrays, matrices,
		other ColMats, or any combination thereof.
		'''
		if len(mats) == 1 :
			mat = mats[0]
			
			if isinstance(mat, ColMat) :
				self.mat = mat.mat #Support a copy constructor.
			else :
				self.mat = np.array(mat) #Simply set self.mat with the numpy array version of the mat.
		elif len(mats) > 1 :
			self.mat = ColMat._mats(*[ColMat(mat) for mat in mats]).mat
		elif not mats :
			self.mat = np.identity(3)
		
	def _mats(*inMats) :
		'''
		Initialize a combined Transform matrix from several input ColMats. Use constructor instead.
		'''
		return ColMat(reduce(ColMat.__mul__, reversed(inMats))) #Works because multiply is actually non-commutative dot.
		#This is why we reverse inMats.
	
	def sample(self, fSeq) :
		shp = np.shape(fSeq)
		if len(shp) == 1 :
			return self.mat.dot(fSeq)
		if len(shp) == 3 :
			#C++ based olo.matr replaces & sped up the operation by 50x with same output!!!
			return olo.matr(fSeq.reshape(reduce(lambda a, b: a*b, fSeq.shape)), self.mat.reshape(reduce(lambda a, b: a*b, self.mat.shape))).reshape(fSeq.shape)
		
	def inv(obj) :
		if isinstance(obj, ColMat) : #Works on any ColMat object - including self.
			return ColMat(np.linalg.inv(obj.mat))
		else : #Works on raw numpy arrays as well.
			return np.linalg.inv(obj)
		
	def transpose(self) :
		return ColMat(np.transpose(self.mat))
		
	#Overloading
	def __mul__(self, other) :
		'''
		* implements matrix multiplication.
		'''
		if isinstance(other, ColMat) :
			return ColMat(self.mat.dot(other.mat))
		elif isinstance(other, float) or isinstance(other, int) :
			return ColMat(np.multiply(self.mat, other))
		elif isinstance(other, np.ndarray) or isinstance(other, np.matrixlib.defmatrix.matrix) :
			return ColMat(self.mat.dot(np.array(other)))
		else :
			raise ValueError('Invalid multiplication arguments!')
			
	__rmul__ = __mul__
	
	def __add__(self, other) :
		if isinstance(other, ColMat) :
			return ColMat(self.mat + other.mat)
		elif isinstance(other, np.ndarray) or isinstance(other, np.matrixlib.defmatrix.matrix) :
			return ColMat(self.mat + np.array(other))
		else :
			raise ValueError('Invalid addition arguments!')
			
	__radd__ = __add__
	
	def __pow__(self, other) :
		'''
		** implements direct multiplication. You usually don't want this.
		'''
		if isinstance(other, ColMat) :
			return ColMat(self.mat * other.mat)
		elif isinstance(other, float) or isinstance(other, int) :
			return ColMat(np.multiply(self.mat, other))
		elif isinstance(other, np.ndarray) or isinstance(other, np.matrixlib.defmatrix.matrix) :
			return ColMat(self.mat * np.array(other))
		else :
			raise ValueError('Invalid multiplication arguments!')
			
	def __invert__(self) :
		return self.inv()
	
	def __len__(self) :
		return len(self.mat)
		
	def __getitem__(self, key) :
		return self.mat[key]
		
	def __iter__(self) :
		return iter(self.mat)
		
	def __repr__(self) :
		return "\nColMat (\n{0}  )\n".format(str(self.mat))
		




#Set it up so that the users don't see the files containing the classes.
from .Transform import Transform
from .ColMap import ColMap
from .LUT import LUT
from .Func import Func
from .ColMat import ColMat
from .Viewer import Viewer

#Ensure the package namespace lines up.
from . import gamma
from . import gamut

__all__ = ['ColMap', 'Transform', 'LUT', 'Func', 'ColMat', 'Viewer', 'gamma', 'gamut']




import abc

import numpy as np

class Transform :
	def spSeq(seq, outLen) :
		"""
		Utility function for splitting a sequence into equal parts, for multithreading.
		"""
		perfSep = (1/outLen) * len(seq)
		return list(filter(len, [seq[round(perfSep * i):round(perfSep * (i + 1))] for i in range(len(seq))])) if len(seq) > 1 else seq
		
	@abc.abstractmethod
	def sample(self, fSeq) :
		"""
		Samples the Transformation.
		"""



#!/usr/bin/env python3

'''
Copyright 2016 Sofus Rose

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
'''

import sys, os, time
import multiprocessing as mp

import numpy as np

#Matplotlib is optional.
MOD_MATPLOTLIB = False
try:
	import matplotlib.pyplot as plt
	import matplotlib.mlab as mlab
	
	MOD_MATPLOTLIB = True
except:
	print("Matplotlib not installed. Graphs won't be drawn")

class Files :
	"""
	The Files object is an immutable sequence of files, which supports writing simultaneously to all the files.
	"""
	def __init__(self, *files) :
		seq=[]
		for f in files:
			if isinstance(f, Files): seq += f.files
			elif 'write' in dir(f): seq.append(f)
			else: raise TypeError('Wrong Input Type: ' + repr(f))

		self.files = tuple(seq) #Immutable tuple of file-like objects,



	def write(self, inStr, exclInd=[]) :
		"""
		Writes inStr all file-like objects stored within the Files object. You may exclude certain entries with a sequence of indices.
		"""
		for f in enumerate(self.files) :
			if f[0] in exclInd: continue
			f[1].write(inStr)




	def __add__(self, o, commut=False) :
		"""
		Implements merging with Files objects and appending file-like objects. Returns new Files object.
		"""
		if isinstance(o, Files) :
			this, other = self.files, o.files
		elif 'write' in dir(o) :
			this, other = self.files, [o]
		else :
			return None

		if commut: this, other = other, this

		return Files(*this, *other) #this and other must be unpackable.

	def __radd__(self, o) :
		"""
		Commutative addition.
		"""
		return self.__add__(o, commut=True) #Use the add operator. It's commutative!!

	def __bool__(self) :
		"""
		False if empty.
		"""
		return bool(self.files)



	def __getitem__(self, index) :
		"""
		Supports slicing and indexing.
		"""
		if isinstance(index, slice) :
			return Files(self.files[index])
		else :
			return self.files[index]

	def __len__(self) :
		"""
		Number of files in the Files object.
		"""
		return len(self.files)

	def __repr__(self) :
		return 'Files(' + ', '.join("'{}'".format(n.name) for n in self.files) + ')'

	def __iter__(self) :
		"""
		Iterates through the file-like objects.
		"""
		return iter(self.files)

class ColLib :
	"""
	Simple hashmap to colors.. Make sure to activate colors in ~/.bashrc or enable ansi.sys!
	"""
	cols = {	'HEADER' : '\033[97m',
				'OKBLUE' : '\033[94m',
				'OKGREEN' : '\033[92m',
				'WARNING' : '\033[93m',
				'FAIL' : '\033[91m',
				'CRIT' : '\033[31m',
				'DEBUG' : '\033[35m',

				'ENDC' : '\033[0m',

				'BOLD' : '\033[1m',
				'ITALIC' : '\033[3m',
				'UNDERLINE' : '\033[4m'
	}



	debug = {	'info' : ('[INFO]', 'OKGREEN'),
				'error' : ('[ERROR]', 'FAIL'),
				'crit' : ('[CRIT]', 'CRIT'),
				'warn' : ('[WARNING]', 'WARNING'),
				'debug' : ('[DEBUG]', 'DEBUG'),
				'run' : ('[RUN]', 'OKBLUE')
	}



	def colString(color, string) :
		"""
		Returns a colored string.
		"""
		return '{}{}{}'.format(cols[color], string, cols['ENDC'])

	def dbgString(signal, string) :
		"""
		"""
		return '{}{}{}'.format(debug[signal])



	def printCol(color, colored, *output, **settings) :
		"""
		Simple print clone where the first printed parameter is colored.
		"""
		print(cols[color] + str(colored) + cols['ENDC'], *output, **settings)

	def printDbg(signal, *output, **settings) :
		"""
		Pass in simple debug signals to print the corresponding entry.
		"""
		printCol(debug[signal][1], debug[signal][0] + ' ' + colored, *output, **settings)

class Log(ColLib) :
	"""
	Logging object, an instance of which is passed throughout afarm. **It has + changes state**, as the sole exception to the
	'no globals' paradigm. You may pass in any file-like object, the only criteria being that it has a 'write' method. stdout is
	used by default.
	"""
	def __init__(self, *file, verb=3, useCol=True, startTime=None) :
		if not file: file = [sys.stdout]
		if startTime is None: startTime = time.perf_counter()

		self.verb = verb #From 0 to 3. 0: CRITICAL 1: ERRORS 2: WARNINGS 3: DEBUG. Info all.
		self.file = Files(*file)

		self.log = [] #Log list. Format: (verb, time in ms, debug_text, text)
		self.sTimes = dict() #Dict of start times for various runs.
		
		self._useCol = useCol #Whether or not to use colored output.
		self._attrLock = mp.Lock() #The log access lock.
		self._startTime = startTime #Global instance time. begins when the instance is created.



	def getLogTime(self) :
		"""
		Gets the current logging time in seconds, from the time of instantiation of the Log object.
		"""
		return time.perf_counter() - self._startTime
		
	def startTime(self, run) :
		"""
		Starts the timer for the specified run. Can use any immutable object to mark the run.
		"""
		self.sTimes[run] = self.getLogTime()
		
	def getTime(self, run) :
		"""
		Gets the time since startTime for the specified run (an immutable object).
		"""
		if run in self.sTimes :
			return self.getLogTime() - self.sTimes[run]
		else :
			raise ValueError('Run wasn\'t found!!')
			
	@staticmethod
	def bench(f, args=[], kwargs={}, trials=15, graph=False) :
		def t(): l = Log(); l.startTime(0); f(*args, **kwargs); return l.getTime(0)
		
		data = np.array([t() for i in range(trials)])
		anyl = {		'mean'		: np.mean(data),
						'median'	: np.median(data),
						'std_dev'	: np.std(data),
						'vari'		: np.std(data) ** 2,
						'total'		: sum(data)
		}
		
		if graph: Log.graphBench(anyl)
			
		return anyl
		
	@staticmethod
	def graphBench(anyl) :
		if MOD_MATPLOTLIB :
			fig = plt.figure()
			
			x = np.linspace(-3 * anyl['std_dev'] + anyl['mean'], 3 * anyl['std_dev'] + anyl['mean'], 100)
			
			plt.plot(x, mlab.normpdf(x, anyl['mean'], anyl['std_dev']))
			
			plt.axvline(x = anyl['mean'], color='red', linestyle = "--")
			plt.text(	anyl['mean'] - 0.2 * anyl['std_dev'], 0, 'mean',
						horizontalalignment = 'left', verticalalignment='bottom',
						rotation = 90, fontsize=10, fontstyle='italic'
			)
			plt.xlabel('Time (Seconds)', fontsize=15)
			plt.ylabel('Distribution', fontsize=11)
			
			plt.show()

	def compItem(self, state, time, *text) :
		"""
		Returns a displayable log item as a string, formatted with or without color.
		"""
		decor = {	'info' : '',
					'error' : '',
					'crit' : ColLib.cols['BOLD'],
					'warn' : '',
					'debug' : ColLib.cols['BOLD'],
					'run' : ColLib.cols['BOLD']
		}[state]
				
		timeCol = {	'info' : ColLib.cols['HEADER'],
					'error' : ColLib.cols['WARNING'],
					'crit' : ColLib.cols['FAIL'],
					'warn' : ColLib.cols['HEADER'],
					'debug' : ColLib.cols['DEBUG'] + ColLib.cols['BOLD'],
					'run' : ColLib.cols['OKGREEN']
		}[state]
						
		if self._useCol :
			return '{3}{5}{0}{4[ENDC]}\t{6}{1:.10f}{4[ENDC]}: {2}'.format(	ColLib.debug[state][0],
																				time,
																				''.join(text),
																				ColLib.cols[ColLib.debug[state][1]],
																				ColLib.cols,
																				decor,
																				timeCol
																		)
		else :
			return '{0} {1:.10f}: {2}'.format(ColLib.debug[state][0], time, ''.join(text))



	def write(self, *text, verb=2, state='info') :
		"""
		Adds an entry to the log file, as well as to the internal structure.

		Possible state values:
		*'info': To give information.
		*'error': When things go wrong.
		*'crit': When things go very wrong.
		*'warn': To let the user know that something weird is up.
		*'run': To report on an intensive process.
		*'debug': For debugging purposes. Keep it at verbosity 3.
		
		Possible verbosity values, and suggested usage:
		*0: User-oriented, general info about important happenings.
		*1: Helpful info about what is running/happening, even to the user.
		*2: Deeper info about the programs functionality, for fixing problems.
		*3: Developer oriented debugging.
		"""
		text = [str(t).strip() for t in text if str(t).strip()]
		if not text: return #Empty write's are no good.
		curTime = self.getLogTime()
		with self._attrLock :
			self.log.append( {	'verb' 	: verb,
								'time' 	: curTime,
								'state'	: state,
								'text'	: ' '.join(str(t) for t in text)
							}
			)

		if self.verb >= verb :
			with self._attrLock :
				print(self.compItem(state, curTime, ' '.join(text)), file=self.file)

	def read(self, verb=None) :
		"""
		Reads the internal logging data structure, optionally overriding verbosity.
		"""
		if not verb: verb = self.verb
		with self._attrLock :
			return '\n'.join([self.compItem(l['state'], l['time'], l['text']) for l in self.log if verb >= l['verb']])

	def reset(self, startTime=None) :
		return Log(self.file, verb=self.verb, useCol=self._useCol, startTime=startTime)



	def getFiles(self) :
		"""
		Get the list of files to dump Log output to.
		"""
		return self.file

	def setFiles(self, *files) :
		"""
		Set a new list of files to dump Log output to.
		"""
		with self._attrLock :
			self.file = Files(*files)

	def addFiles(self, *files) :
		"""
		Add a list of files to dump Log output to.
		"""
		with self._attrLock :
			self.file += Files(*files)



	def setVerb(self, newVerb) :
		"""
		Call to set verbosity.
		"""
		with self._attrLock :
			self.verb = newVerb
		self.write('Verbosity set to', str(newVerb) + '.', verb=0, state='info')

	def setCol(self, newUseCol) :
		"""
		Call to change color output.
		"""
		with self._attrLock :
			self._useCol = newUseCol
		self.write('Color Output set to', self._useCol, verb=0, state='info')



	def __call__(self, verb, state, *text) :
		"""
		Identical to Log.write(), except it requires the verbosity level to be specified.
		"""
		self.write(verb=verb, state=state, *text)

	def __repr__(self) :
		return (	'Log(' +
					(', '.join("'{}'".format(f.name) for f in self.file.files) + ', ' if self.file else '') +
					'verb={}, useCol={}, startTime={:.3f})'.format(self.verb, self._useCol, self._startTime)
			)

	def __str__(self) :
		return self.read()



	def __add__(self, o, commut=False) :
		"""
		Merges a Log object with another Log object, a Files object, or a file-like object.
		*For Log object addition, the minimum startTime attibute is used to initialize the merged startTime.
		*The Files objects of both Log objects are merged.
		"""

		if isinstance(o, Log) :
			l = self.reset(min(self.getLogTime(), o.getLogTime())) #Max of self and other time.
			l.log = self.log + o.log
			l.log.sort(key=lambda item: item['time']) #Make sure to sort the internal log by time.
			l.addFiles(o.file)
			return l
		elif isinstance(o, Files) :
			l = self.reset(self.getLogTime())
			l.log = self.log
			l.setFiles(self.getFiles(), *o.files)
			return l
		elif 'write' in dir(o) :
			l = self.reset(self.getLogTime())
			l.file = o + self.file if commut else self.file + o
			return l
		else :
			return None

	def __radd__(self, o) :
		return self.__add__(o, commut=True) #Use the add operator. It's commutative!!


	def __bool__(self) :
		"""
		False if log is empty.
		"""
		return bool(self.log)

	def __getitem__(self, i) :
		"""
		Supports slicing and indexing, from recent (0) to oldest (end).
		"""
		if isinstance(i, slice) :
			l = self.reset(self.getLogTime())
			for ind, itm in enumerate(self.log) :
				if ind in list(range(i.start if i.start else 0, i.stop, i.step if i.step else 1)): l.log.append(itm)
			return list(l)
		else :
			l = self.log[::-1][i]
			return self.compItem(l['state'], l['time'], l['text'], noCol = not self._useCol)

	def __len__(self) :
		"""
		Amount of items in the log.
		"""
		return len(self.log)

	def __iter__(self) :
		"""
		Iterator never colors output.
		"""
		return iter(self.compItem(l['state'], l['time'], l['text'], noCol = True) for l in self.log)

class LogFile() :
	"""
	Similar to a normal file, except it splits into several files. On the frontend, however, it acts as if it were a single file.
	
	*Writes to 'path'.log.
	*When maxLen is exceeded, lines are pushed into 'path'.0.log, then 'path'.1.log, etc. .
	"""
	def __init__(self, path, maxLen=1000, trunc=False) : #Make maxLen 1000 later.
		"""
		Constructor accepts a path (extension will be rewritten to '.log'), a maximum length, and will optionally truncate
		any previous logfiles.
		"""
		self.path = os.path.splitext(path)[0] + '.log'
		self.bPath = os.path.splitext(self.path)[0]
		
		self.maxLen = maxLen
		self.name = '<{0}.log, {0}.0...n.log>'.format(self.bPath)
		
		self.lines = 0
		self.fileNum = 0
		
		#If the logfile already exists, it's read + rewritten using the current maxLen.
		if os.path.exists(self.path) :
			if trunc: self.truncate(); return
			
			inLines = open(self.path, 'r').readlines()[::-1]
			os.remove(os.path.abspath(self.path)) #Remove the old path.log.
			
			i = 0
			while os.path.exists('{0}.{1}.log'.format(self.bPath, i)) :
				inLines += open('{0}.{1}.log'.format(self.bPath, i), 'r').readlines()[::-1]
				os.remove(os.path.abspath('{0}.{1}.log'.format(self.bPath, i)))
				i += 1
							
			self.write(''.join(reversed(inLines)))
			
	
	def write(self, *inStr) :
		apnd = list(filter(bool, ''.join(inStr).strip().split('\n')))
		
		if not apnd: return #Nothing to append = don't even try!
		if not os.path.exists(self.path): open(self.path, 'w').close() #Make sure path.log exists.
				
		#Empty apnd line by line.
		while len(apnd) > 0 :
			toWrite = self.maxLen * (self.fileNum + 1) - self.lines #Lines needed to fill up path.log
			if toWrite == 0 : #Time to make new files.		
						
				#Rename upwards. path.n.log -> path.(n+1).log, etc. . path.log becomes path.0.log.
				for i in reversed(range(self.fileNum)) :
					os.rename('{0}.{1}.log'.format(self.bPath, i), '{0}.{1}.log'.format(self.bPath, i+1))
				os.rename('{0}.log'.format(self.bPath), '{0}.0.log'.format(self.bPath))
				
				#Make new path.log.
				open(self.path, 'w').close() #Just create the file.
				
				#Number of files just increased.
				self.fileNum += 1
			else : #Fill up path.log.
				print(apnd.pop(0), file=open(self.path, 'a'))
				
				#Number of written lines just increasd.
				self.lines += 1
		
	def read(self) :
		collec = []
		
		for i in reversed(range(self.fileNum)) :
			collec += open('{0}.{1}.log'.format(self.bPath, i), 'r').readlines()
			
		collec += open(self.path, 'r').readlines()
		
		return ''.join(collec)
		
	def truncate(self) :
		"""
		Deletes all associated files + resets the instance.
		"""
		i = 0
		
		os.remove(os.path.abspath(self.path)) #Remove the old path.log.
		while os.path.exists('{0}.{1}.log'.format(self.bPath, i)) : #Remove all log files.
			os.remove(os.path.abspath('{0}.{1}.log'.format(self.bPath, i)))
			i += 1
			
		self.lines = 0
		self.fileNum = 0
		
	def readlines(self) :
		return self.read().split('\n')
		
	def isatty(self) :
		"""
		Always returns false, as a LogFile is never associated with a tty.
		"""
		return False
		
	def __iter__(self) :
		return (line for line in self.readlines())
		
	def __repr__(self) :
		return 'LogFile({0}, maxLen={1})'.format(self.path, self.maxLen)

def coolTest() :
	l = Log(Files(LogFile('first', 10, True), LogFile('second', 20, True)), LogFile('third', 30, True), sys.stdout)
	l(0, 'info', 'Big Failure Oh NO!')
	l(0, 'error', 'Big Failure Oh NO!')
	l(0, 'crit', 'Big Failure Oh NO!')
	l(0, 'warn', 'Big Failure Oh NO!')
	l(0, 'debug', 'Big Failure Oh NO!')
	l(0, 'run', 'Big Failure Oh NO!')
	
	print('We got ourselves a log file here kids.')
	
	for i in range(50) :
		l(0, 'run', 'This is the', i, 'run today!')
	
	print(l.getLogTime())

def logFileTest() :
	l = LogFile('hi.log', maxLen = 3, trunc=True)
	
	print('\n', repr(l), sep='')
	print('hi', 'world', file=l, sep='\n')
	
	print('\n', repr(l), sep='')
	print('you', 'are', 'cool', 'friend', 'hi', file=l, sep='\n')
	print('hello\nmotherfucker\nits\narnold\nyour\nold\nfriend\nrunbaby!!\nlittlebitch :)', file=l)
	print('\n', repr(l), sep='')
	print('Reading Log Files:', repr(l.read()))
	print(l.name, '\n\n\n')
	
	print(l.read().split('\n'))
	print(len([x for x in l.read().split('\n') if x]))
	
	l = LogFile('hi.log', 10)
	print(l.read().split('\n'))
	print(len([x for x in l.read().split('\n') if x]))
		
	l = Log(LogFile('hi', 100, False))
	print(repr(l))
	for x in range(10) :
		l(0, x)
	#~ l.setCol(False)
	l(0, 'info', 'Big Failure Oh NO!')
	l(0, 'error', 'Big Failure Oh NO!')
	l(0, 'crit', 'Big Failure Oh NO!')
	l(0, 'warn', 'Big Failure Oh NO!')
	l(0, 'debug', 'Big Failure Oh NO!')
	l(0, 'run', 'Big Failure Oh NO!')
	l = LogFile('hi', 500, False)
	print('hihi', 'you', 'can\'t', 'beat', 'the', 'trunc', file=l, sep='\n')

def logTest() :
		l = Log(open('hi.txt', 'w'))
		#~ l.setCol(False)
		print(repr(l))
		print('1', file=l)
		l(1, '2')
		a = l.reset()

		#~ l.setCol(True)

		l.addFiles(sys.stderr)
		print('\n', repr(l), sep='')

		print('\n', repr(a), sep='')
		a(2, '3')
		a.write('4')
		a.addFiles(sys.stdout)

		print('\n a + l ', repr(a + l), sep='')

		print(a + l)

		print('\n', repr(a + sys.stderr), sep='')
		print('\n', repr(sys.stderr + a), sep='')
		print('\n', (l + a), sep='')
		print('\n', (l + a)[0:3], sep='')

		for item in (l + a) :
			print(item)

		print("\nLength of l + a: ", len(l + a))
		print("\nl + a: ", repr(Log() + Files(sys.stdout)))
		print('\n', l.read(verb=0), sep='')

if __name__ == "__main__" :
	#~ unitTest()
	#~ logFileTest()
	coolTest()



import sys, os, os.path

from functools import reduce

import numpy as np

import wand
import wand.image
import wand.display

from wand.api import library

#~ library.MagickSetCompressionQuality.argtypes = [ctypes.c_void_p, ctypes.c_size_t]
#~ library.MagickSetCompression.argtypes = [ctypes.c_void_p, ctypes.c_size_t]

#~ COMPRESS_TYPES = dict(zip(wand.image.COMPRESSION_TYPES, tuple(map(ctypes.c_int, range(len(wand.image.COMPRESSION_TYPES))))))

from . import gamma
from .LUT import LUT
from .Viewer import Viewer

from .lib import olOpt as olo

class ColMap :
	'''
	The ColMap class stores an image in its 32 bit float internal working space.
	
	:var DEPTHS: A dictionary of depths in relation to the Depths dictionary.
	
	ColMaps are initialized by default with 0's; a black image. You can use
	`open` to load a path, :py:func:`~fromArray` to load from a numpy array, or :py:func:`~fromBinary` to load from
	a binary representation (useful in pipes).
	
	:param shape: The numpy-style shape of the empty image. Specify width, then height.
	:type shape: tuple[int, int] or tuple[int, int, int]
	:param depth: The integer depth used for int format's input and output. Set to DEPTHS['full'] by default.
	:type depth: int or None
	
	:return: An empty ColMap, holding a black image of specified shape.
	:raises ValueError: When trying to use unsupported bit depth.
	:raises ValueError: When using invalid image shape.
	'''
	
	DEPTHS = {	'default'	: None,
				'comp'		: 8,
				'half'		: 16,
				'full'		: 32,
				'double'	: 64
	}
	
#Constructors
	def __init__(self, shape, depth = None) :
		if depth not in ColMap.DEPTHS.values() :
			raise ValueError('Bit depth not supported! Supported bit depths: {}'.format(', '.join(ColMap.DEPTHS.values())))
		
		if len(shape) not in (2, 3) :
			raise ValueError('Please use a valid numpy image array shape!')
		
		self.depth = depth if depth is not None else ColMap.DEPTHS['full'] #This represents the real precision of data.
		self.rgbArr = np.zeros((shape[0], shape[1], 3), dtype=np.float32)
		
	@staticmethod
	def fromArray(imgArr) :
		'''
		Initialize a ColMap from a numpy array of either float or int type (containing an image).
		
		See :py:class:`~ColMap` initialization for a lower-level constructor.
		
		:param imgArr: The numpy image array. Must have shape (width, height, 3)
		:param depth: The integer depth used for int format's input and output. None will use highest available.
		:type depth: int or None
		
		:return: A ColMap containing the image represented in imgArr.
		:raises ValueError: When trying to use unsupported array data type
		'''
		
		#Infer bitDepth from array to create new array, nArr, which we'll use to make our ColMap.
		
		if issubclass(imgArr.dtype.type, np.integer) : #If it's an integer.
			bitDepth = int(''.join([i for i in str(imgArr.dtype) if i.isdigit()]))
			
			nArr = np.divide(imgArr.astype(np.float32), 2 ** bitDepth - 1)
			
		elif issubclass(imgArr.dtype.type, np.floating) : #It it's a float.
			#If we're dealing with an np.float16 array, we can't exactly start giving 32 bit output.
			if int(''.join([i for i in str(imgArr.dtype) if i.isdigit()])) == 16 :
				bitDepth = 16
			else :
				bitDepth = None
				
			nArr = np.array(imgArr, dtype=np.float32)
			
		else :
			raise ValueError('The input image array uses an invalid data type {}! Please use any np.int or np.float variant!'.format(imgArr.dtype.type))
		
				
		#We're taking over the creation of img.rgbArr, so we need to do different error checking of our own.
		if len(nArr.shape) not in (2, 3) :
			raise ValueError('Please use a valid numpy image array shape!')
			
		elif len(nArr.shape) == 2 :
			
			#If we're dealing with a greyscale image, then we need to convert it to RGB using an optimized C++ function.
			nArr = olo.grey_to_rgb(nArr.reshape(reduce(lambda a, b: a*b, nArr.shape))).reshape((nArr.shape[0], nArr.shape[1], 3))
		
		img = ColMap(nArr.shape, depth=bitDepth)
		img.rgbArr = nArr
		
		return img
		
	@staticmethod
	def fromBinary(binData, fmt, width=None, height=None) :
		'''
		Construct a ColMap from an image in binary form. See :py:func:`~ColMap.toBinary` for the inverse.
		
		* This won't work for greyscale data - it's assumed to be RGB.
		
		:param bin binData: The binary data blob to open.
		:param str fmt: Wand needs to know what image format the binary data being thrown at it is in! See https://www.imagemagick.org/script/formats.php .
		:param str width: You may specify a specific width if you're having problems.
		:param str height: You may specify a specific height if you're having problems.
		:return: The image, as a ColMat.
		:rtype: :py:class:`~ColMap`
		
		This is great for pipes, where you're receiving binary data through stdin.
			* Set binData to `sys.stdin.buffer.read()` in a script to pipe data into it!
			
		**NOTE: Uses Wand's "blob" functionality, and as such incurs Wand's limitations.**
		'''
		with wand.image.Image(blob=binData, format=fmt, width=width, height=height) as img:
			
			return ColMap.fromArray(np.fromstring(img.make_blob("RGB"), dtype='uint{}'.format(img.depth)).reshape(img.height, img.width, 3))
	
		
	@staticmethod
	def open(path) :
		'''
		Construct a ColMap from an image on the disk.
		
		:param str path: The image path to open.
		:return: The image, as a ColMat.
		:rtype: :py:class:`~ColMap`
		
		ColMap currently uses ImageMagick to open a wide range of formats, including:
		
		* **EXR**: The industry standard for HDR, wide-gamut, linear-encoded images.
		* **DPX**: An older production format.
		* **PNG**: Can store 16-bit images well. Usually quite slow.
		* *Any other IM-supported formats...* See https://www.imagemagick.org/script/formats.php
		'''
		
		try :
			openFunction = {
				"exr" : ColMap.openWand,
				"dpx" : ColMap.openWand,
			}[path[path.rfind('.') + 1:]]
			
			return openFunction(path) #Any fancy formats will go here.
		except :
			#Fallback to opening using Wand.
			return ColMap.openWand(path)
		
#Operations - returns new ColMaps.
	def apply(self, transform) :
		'''
		Apply an image transformation, in the form of a subclass of :py:class:`~Transform`.
		
		You can apply LUTs, gamma functions, matrices - simply insert an instance of :py:class:`~LUT`,
		:py:class:`~Func`, :py:class:`~ColMat`, or any other :py:class:`~Transform` object to apply it
		to the image!
		
		:param transform: An image transform.
		:type transform: :py:class:`~Transform`
		:return: A transformed ColMap.
		'''
		return ColMap.fromArray(transform.sample(self.asarray()))
		
#Vendor-specific open methods.
	@staticmethod
	def openWand(path) :
		'''
		Vendor-specific :py:func:`~ColMap.open` function. See :py:func:`~ColMap.open`
		
		:param str path: The image path to open.
		:return: The image, as a ColMat.
		:rtype: :py:class:`~ColMap`
		'''
		
		with wand.image.Image(filename=path) as img:
			#Quick inverse sRGB transform, to undo what Wand did - but not for exr's, which are linear bastards.
			if img.format != 'EXR' :
				img.colorspace = 'srgb'
				img.transform_colorspace('rgb')
			
				img.colorspace = 'srgb' if img.format == 'DPX' else 'rgb' #Fix for IM's dpx bug.
			
			return ColMap.fromArray(np.fromstring(img.make_blob("RGB"), dtype='uint{}'.format(img.depth)).reshape(img.height, img.width, 3))
	
	def save(self, path, compress = None, depth = None) :
		'''
		Save a ColMap to an image file on the disk.
		
		:param str path: The path to save the image file at. The extension specified determines the output format.
		:param compress: Compression options passed to the vendor. Currently broken.
		:type compress: str or None
		:param depth: You may override the ColMap's depth if you wish.
		:type depth: int or None
		
		
		ColMap currently uses ImageMagick to save a wide range of formats, including:
		
		* **EXR**: The industry standard for HDR, wide-gamut, linear-encoded images.
		* **DPX**: An older production format.
		* **PNG**: Can store 16-bit images well. Usually quite slow.
		* *Any other IM-supported formats...* See https://www.imagemagick.org/script/formats.php
		
		**NOTE: EXRs are only saveable as 16-bit integer, with no compression options. This is an IM/Wand library limitation.**
		'''
		
		if depth not in ColMap.DEPTHS.values() :
			raise ValueError('Bit depth not supported! Supported bit depths: {}'.format(', '.join(ColMap.DEPTHS.values())))
		
		try :
			saveFunction = {
				"exr" : self.saveWand,
				"dpx" : self.saveWand,
				"tif" : self.saveWand,
				"tiff": self.saveWand
			}[path[path.rfind('.') + 1:]]
			
			return saveFunction(path, compress, depth)
		except :
			#Fallback to saving using Wand.
			self.saveWand(path, compress, depth)
			

#Vendor-specific save methods

	def saveWand(self, path, compress = None, depth = None) :
		'''
		Vendor-specific :py:func:`~ColMap.save` function. See :py:func:`~ColMap.save`
		
		:param str path: The image path to save to.
		:param compress: Compression options passed to Wand. Currently broken.
		:param depth: You may override the ColMap's depth if you wish.
		:type depth: int or None
		
		**NOTE: EXRs are only saveable as 16-bit integer, with no compression options. This is an IM/Wand library limitation.**
		'''
		
		if depth not in ColMap.DEPTHS.values() :
			raise ValueError('Bit depth not supported! Supported bit depths: {}'.format(', '.join(ColMap.DEPTHS.values())))
		
		data = self.apply(LUT.lutFunc(gamma.sRGB)) if path[path.rfind('.')+1:] == 'dpx' else self
		i = data.asWandImg(depth)
		
		i.colorspace = 'srgb' #Make sure it saves without a colorspace transformation.
		
		#~ if compress :
			#~ library.MagickSetCompression(i.wand, 'rle')
			
			#~ i.compression = 'lzma'
			#~ i.compression_quality = 80

		i.save(filename=path)
		

#Display Functions
	
	@staticmethod
	def display(path, width = 1000) :
		'''
		Display an image at a path on the disk, using the builtin OpenGL Viewer.
		
		:param width: The desired width of the viewer; the height is automatically gleaned from the aspect ratio.
		
		For the viewer source code, see :py:class:`~Viewer`.
		'''
		
		img = ColMap.open(path).rgbArr
		aspectRatio = img.shape[0]/img.shape[1]
			
		xRes = width
		yRes = int(xRes * aspectRatio)
		
		Viewer.run(img, xRes, yRes, title = os.path.basename(path))
		
	def show(self, width = 1000) :
		'''
		Display this ColMap using the builtin OpenGL Viewer.
		
		:param width: The desired width of the viewer; the height is automatically gleaned from the aspect ratio.
		
		For the viewer source code, see :py:class:`~Viewer`.
		'''
		
		#Use my custom OpenGL viewer!
		Viewer.run(self.rgbArr, width, int(width * self.rgbArr.shape[0]/self.rgbArr.shape[1]))
		

#Data Output Types
	def asWandImg(self, depth = None) :
		'''
		Output this ColMap as a Wand image.
				
		:param depth: You may override the ColMap's depth if you wish.
		:type depth: int or None
		:return: The Wand Image.
		:rtype: wand.image
		
		See http://docs.wand-py.org/en/0.4.4/index.html for Wand docs.
		'''
		
		if depth not in ColMap.DEPTHS.values() :
			raise ValueError('Bit depth not supported! Supported bit depths: {}'.format(', '.join(ColMap.DEPTHS.values())))
		
		if depth is None :
			d = ColMap.DEPTHS['half'] if self.depth >= ColMap.DEPTHS['half'] else self.depth #Highest is half - 16.
		else :
			d = depth
		
		#~ i = wand.image.Image(blob=self.asarray().tostring(), width=np.shape(self.rgbArr)[1], height=np.shape(self.rgbArr)[0], format='RGB') #Float Array
		i = wand.image.Image(blob=self.asIntArray(d).tostring(), width=np.shape(self.rgbArr)[1], height=np.shape(self.rgbArr)[0], format='RGB')
		i.colorspace = 'rgb' #Specify, to Wand, that this image is to be treated as raw, linear, data.
		
		return i
		
			
	def toBinary(self, fmt, depth=None) :
		'''
		Output this ColMap in binary form. See :py:func:`~ColMap.fromBinary` for the inverse.
				
		:param str fmt: Wand needs to know what format to output! See https://www.imagemagick.org/script/formats.php .
		:param depth: You may override the ColMap's bit depth if you wish.
		:type depth: int or None
		:return: The image, as a ColMat.
		:rtype: :py:class:`~ColMap`
		
		This is great for pipes, where you're sending binary data through stdout.
			* Use the return value as the argument of sys.stdout.write() to pipe the image to other applications!
			
		**NOTE: Uses Wand's "blob" functionality, and as such incurs Wand's limitations.**
		'''
		
		if depth not in ColMap.DEPTHS.values() :
			raise ValueError('Bit depth not supported! Supported bit depths: {}'.format(', '.join(ColMap.DEPTHS.values())))
		
		with self.asWandImg(d) as img :
			img.format = fmt
			return img.make_blob()
		
	def asarray(self) :
		"""
		Returns the internal np.float32 image array directly.
		
		:return: The internal numpy array.
		:rtype: np.array
		"""
		return self.rgbArr
		
	def asIntArray(self, depth = None, us = True) :
		"""
		Returns the internal image array as an int array.
		
		:param depth: You may override the ColMap's bit depth if you wish.
		:type depth: int or None
		:param bool us: True will output unsigned ints, False will output signed ints.
		:return: The internal numpy array.
		:rtype: np.array
		"""
		
		if depth not in ColMap.DEPTHS.values() :
			raise ValueError('Bit depth not supported! Supported bit depths: {}'.format(', '.join(ColMap.DEPTHS.values())))
		
		if depth is None :
			d = self.depth #No limits here.
		else :
			d = depth
		
		u = 'u' if us else '' #Unsigned or no?
		return np.multiply(self.rgbArr.clip(0, 1), 2.0 ** depth - 1).astype("{0}int{1}".format(u, depth))
		
		
#Overloads
	def __repr__(self) :
		return 'ColMap.fromArray( \n\trgbArr = {0}\n)'.format('\n\t\t'.join([line.strip() for line in repr(self.rgbArr).split('\n')]))



import multiprocessing as mp
import types
from functools import reduce

import numpy as np

from .Transform import Transform
from .lib import olOpt as olo

class Func(Transform) :
	def __init__(self, func) :
		self.func = func
		
	#Func Methods
	def __gamma(q, cpu, f, spSeq) :
		q.put( (cpu, f(spSeq)) )
	
	def sample(self, fSeq) :
		fSeq = np.array(fSeq, dtype=np.float32) #Just some type assurances.
		
		# Any float-returning C++ functions can be threaded with olo.gam(), but because of GIL, it won't work with Python functions.
		if isinstance(self.func, types.BuiltinFunctionType) :
			# \/ Just olo.gam, except fSeq is flattened to a 1D array, processed flat, then shaped back into a 3D array on the fly.
			return olo.gam(fSeq.reshape(reduce(lambda a, b: a*b, fSeq.shape)), self.func).reshape(fSeq.shape) #OpenMP vectorized C++ motherfuckery!
		else :
			#We always have the slow af fallback.
			fVec = np.vectorize(self.func)
			
			out = []
			q = mp.Queue()
			splt = Transform.spSeq(fSeq, mp.cpu_count())
			for cpu in range(mp.cpu_count()) :
				p = mp.Process(target=Func.__gamma, args=(q, cpu, fVec, splt[cpu]))
				p.start()
				
			for num in range(len(splt)) :
				out.append(q.get())
				
			return np.concatenate([seq[1] for seq in sorted(out, key=lambda seq: seq[0])], axis=0) if len(fSeq) > 1 else self.func(fSeq[0])
			
			#~ return fVec(fSeq) if len(fSeq) > 1 else self.func(fSeq[0])
		



#!/usr/bin/env python3

import sys, os
import os.path as path
from os.path import join

from setuptools import setup
from setuptools import Extension
from setuptools import find_packages

#pybind11, a dep of openlut, can report its own includes for gcc compilation!
import pybind11

def read(fname):
    return open(join(path.dirname(__file__), fname)).read()
    
pypi_deps = ['numpy', 'wand', 'scipy', 'pygame','PyOpenGL', 'setuptools', 'pybind11', 'wheel']

#Make sure we're using gcc.
os.environ["CC"] = "g++"
os.environ["CXX"] = "g++"

cpp_args = ['-fopenmp', '-std=gnu++14', '-O3']
link_args = ['-fopenmp']

olOpt = Extension(	'openlut.lib.olOpt',
					sources = ['openlut/lib/olOpt.cpp'],
					include_dirs=[pybind11.get_include()], #Include pybind11 from its pip package.
					language = 'c++',
					extra_compile_args = cpp_args,
					extra_link_args = cpp_args
		)

setup(	name = 'openlut',
		version = '0.2.6',
		description = 'OpenLUT is a practical color management library.',
		long_description = read('README.rst'),
		
		author = 'Sofus Rose',
		author_email = 'sofus@sofusrose.com',
		url = 'https://www.github.com/so-rose/openlut',
		
		packages = find_packages(exclude=['src']),
		
		ext_modules = [olOpt],

		license = 'MIT Licence',
		
		keywords = ['color', 'image', 'images', 'processing'],
		
		install_requires = pypi_deps,
		
		classifiers = [
			'Development Status :: 3 - Alpha',
			
			'License :: OSI Approved :: MIT License',
			
			'Programming Language :: Python :: 3'
		]
)



import multiprocessing as mp
from functools import reduce
import types

import numpy as np

#scipy is an optional dependency.
MOD_SCIPY = False
try :
	from scipy.interpolate import splrep, splev
	MOD_SCIPY = True
except :
	pass

from .Transform import Transform
from .lib import olOpt as olo

class LUT(Transform) :
	def __init__(self, dims = 1, size = 4096, title = "openlut_LUT", iRange = (0.0, 1.0)) :	
		'''
		Create an identity LUT with given dimensions (1 or 3), size, and title.
		'''	
		if dims != 1 and dims != 3: raise ValueError("Dimensions must be 1 or 3!")
		
		self.title = title #The title.
		self.size = size #The size. 1D LUTs: size numbers. 3D LUTs: size x size x size numbers.
		self.range = iRange #The input range - creates data or legal LUTs. Should work fine, but untested.
		self.dims = dims #The dimensions. 1 or 3; others aren't accepted.
		self.ID = np.linspace(self.range[0], self.range[1], self.size, dtype=np.float32) #Read Only.
		
		if dims == 1 :
			self.array = np.linspace(self.range[0], self.range[1], self.size, dtype=np.float32) #Size number of floats.
		elif dims == 3 :
			print("3D LUT Not Implemented!")
			#~ self.array = np.linspace(self.range[0], self.range[1], self.size**3).reshape(self.size, self.size, self.size) #Should make an identity size x size x size array.
		
	def lutFunc(func, size = 4096, dims = 1, title="openlut_FuncGen", iRange = (0.0, 1.0)) :
		'''
		Creates a LUT from a simple function.
		'''
		if dims == 1 :
			lut = LUT(dims=dims, size=size, title=title, iRange=iRange)
			
			#Use fast function sampling if the function is a C++ function.
			vFunc = lambda arr: olo.gam(arr, func) if isinstance(func, types.BuiltinFunctionType) else np.vectorize(func, otypes=[np.float32])
			lut.array = vFunc(lut.array)
			
			return lut
		elif dims == 3 :
			print("3D LUT Not Implemented!")
			
	def lutArray(array, title="Array_Generated") :
		'''
		Creates a LUT from a float array. Elements must be in range [0, 1].
		'''
		if len(np.shape(array)) == 1 :
			lut = LUT(dims=1, size=len(array), title=title)
			lut.array = array
			
			return lut
		elif len(np.shape(array)) == 3 :
			print("3D LUT Not Implemented!")
		else :
			raise ValueError("lutArray input must be 1D or 3D!")
			
	def lutMapping(idArr, mapArr, title="Mapped_Array") :
		'''
		Creates a 1D LUT from a nonlinear mapping. Elements must be in range [0, 1].
		'''
		return LUT.lutArray(splev(np.linspace(0, 1, num=len(idArr)), splrep(idArr, mapArr)))
		
#Transform Functions.
	def _splInterp(q, cpu, spSeq, ID, array) :
		q.put( (cpu, splev(spSeq, splrep(ID, array))) ) #Spline Interpolation. Pretty quick, considering.
	
	def sample(self, fSeq, spl=True) :
		'''
		Apply the 1D LUT to the numpy image array, using fast C++ math.
		
		Latest Performance:
			apply(ol.LUT): 0.026462205679908948,, (avg. 100 Trials) *sRGB LUT
			
		:return: Returns a numpy array with identical shape to the input array.
		'''
				
		fSeq = np.array(fSeq)
		if self.dims == 1 :
			
			#Scipy must be loaded & the LUT must be rediculously small before spline interpolation sets in.
			if (not MOD_SCIPY) or self.size > 25 :
				return olo.lut1dlin(fSeq.reshape(reduce(lambda a, b: a*b, fSeq.shape)), self.array, self.range[0], self.range[1]).reshape(fSeq.shape)
				
			else :
				#~ return np.interp(spSeq, self.ID, self.array) #non-threaded way.
				out = []
				q = mp.Queue()
				splt = Transform.spSeq(fSeq, mp.cpu_count())
				for cpu in range(mp.cpu_count()) :
					p = mp.Process(target=LUT._splInterp, args=(q, cpu, splt[cpu], self.ID, self.array))
					p.start()
					
				for num in range(len(splt)) :
					out.append(q.get())
					
				return np.concatenate([seq[1] for seq in sorted(out, key=lambda seq: seq[0])], axis=0)
			
		elif self.dims == 3 :
			print("3D LUT Not Implemented!")
			
#LUT Functions
	def resized(self, newSize) :
		'''
		Return the LUT, resized to newSize.
		
		1D LUTs: If the new size is lower, we use Linear interpolation. If the new size is higher, we use Spline interpolation.
		* If the current size is too low, use spline regardless.
		'''
		if newSize == self.size: return self
			
		fac = newSize / self.size
		
		useSpl = self.size < newSize #If the new size is lower, we use Linear interpolation. If the new size is higher, we use Spline interpolation.
		if self.size < 128: useSpl = True #If the current size is too low, use spline regardless.
		
		if self.dims == 1 :
			newID = np.linspace(self.range[0], self.range[1], newSize)
			return LUT.lutArray(self.sample(newID, spl=useSpl), title="Resized to {0}".format(newSize))
		if self.dims == 3 :
			print("3D LUT Not Implemented")
			
	def inverted(self) :
		'''
		Return the inverse LUT.
		'''
		return LUT.lutArray(splev(np.linspace(self.range[0], self.range[1], num=self.size), splrep(self.array, np.linspace(self.range[0], self.range[1], num=self.size))))
	
#IO Functions.
				
	def open(path) :
		'''
		Opens any supported file format, located at path.
		'''
		openFunction = {
			"cube" : LUT.openCube,
		}[path[path.rfind('.') + 1:]]
		
		return openFunction(path)
			
	def openCube(path) :
		'''
		Opens .cube files. They must be saved with whitespaces. Referenced by open().
		'''
		lut = LUT() #Mutable luts are not reccommended for users.
	
		with open(path, 'r') as f :
			i = 0
			for line in f :
				#~ if not line.strip(): continue
				sLine = line.strip()
				if not sLine: continue
				
				if sLine[0] == '#': continue
				
				index = sLine[:sLine.find(' ')]
				data = sLine[sLine.find(' ') + 1:]
				
				if index == "TITLE": lut.title = data.strip('"'); continue
				if index == "LUT_1D_SIZE": lut.dims = 1; lut.size = int(data); continue
				if index == "LUT_3D_SIZE": lut.dims = 3; lut.size = int(data); continue
				
				if index == "LUT_1D_INPUT_RANGE": lut.range = (float(data[:data.find(' ')]), float(data[data.find(' ') + 1:])); continue
								
				if lut.dims == 1 and sLine[:sLine.find(' ')] :
					lut.array[i] = float(sLine[:sLine.find(' ')])
					i += 1
				elif lut.dims == 3 :
					print("3D LUT Not Implemened!")
				
		return lut
		
	def save(self, path) :
		'''
		Method that saves the LUT in a supported format, based on the path.
		'''
		saveFunction = {
			"cube" : self.saveCube,
			
		
		}[path[path.rfind('.') + 1:]]
		
		saveFunction(path)
		
	def saveCube(self, path) :
		with open(path, 'w') as f :
			print('TITLE', '"{}"'.format(self.title), file=f)
			
			if self.dims == 1 :
				print('LUT_1D_SIZE', '{}'.format(self.size), file=f)
				print('LUT_1D_INPUT_RANGE', '{0:.6f} {1:.6f}'.format(*self.range), file=f)
				print('# Created by openlut.\n', file=f)
				
				for itm in self.array :
					entry = '{0:.6f}'.format(itm)
					print(entry, entry, entry, file=f)
			elif self.dims == 3 :
				print("3D LUT Not Implemented!")
		
#Overloaded functions
	
	def __iter__(self) :
		if dims == 1 :
			return iter(self.array)
		elif dims == 3 :
			iArr = self.array.reshape(self.dims, self.size / self.dims) #Group into triplets.
			return iter(iArr)
		
	def __getitem__(self, key) :
		return self.sample(key)
		
	def __repr__(self) :
		return 'LUT(\tdims = {0},\n\tsize = {1},\n\ttitle = "{2}"\n\tarray = {3}\n)'.format(self.dims, self.size, self.title, '\n\t\t'.join([line.strip() for line in repr(self.array).split('\n')]))



import math

import numpy as np

from .lib import olOpt as olo

#Static Gamma Functions, borrowed from olo.
#inv goes from space to lin.

lin = olo.lin
sRGB = olo.sRGB
sRGBinv = olo.sRGBinv
Rec709 = olo.Rec709
ReinhardHDR = olo.ReinhardHDR
sLog = olo.sLog
sLog2 = olo.sLog2
DanLog = olo.DanLog

class PGamma :
	'''
	Static class containing python versions of the C++ gamma functions.
	'''
	
	def lin(x): return x

	def sRGB(x) :
		'''
		sRGB formula. Domain must be within [0, 1].
		'''
		return ( (1.055) * (x ** (1.0 / 2.4)) ) - 0.055 if x > 0.0031308 else x * 12.92
	def sRGBinv(x) :
		'''
		Inverse sRGB formula. Domain must be within [0, 1].
		'''
		return ((x + 0.055) / 1.055) ** 2.4 if x > 0.04045 else x / 12.92
		
	def Rec709(x) :
		'''
		Rec709 formula. Domain must be within [0, 1].
		'''
		return 1.099 * (x ** 0.45) - 0.099 if x >= 0.018 else 4.5 * x
		
	def ReinhardHDR(x) :
		'''
		Reinhard Tonemapping formula. Domain must be within [0, 1].
		'''
		return x / (1.0 + x)

	def sLog(x) :
		'''
		sLog 1 formula. Domain must be within [0, 1]. See https://pro.sony.com/bbsccms/assets/
		files/mkt/cinema/solutions/slog_manual.pdf .
		'''
		return ( 0.432699 * math.log(x + 0.037584, 10.0) + 0.616596) + 0.03
		
	def sLog2(x) :
		'''
		sLog2 formula. Domain must be within [0, 1]. See https://pro.sony.com/bbsccms/assets/files/micro/dmpc/training/S-Log2_Technical_PaperV1_0.pdf .
		'''
		return ( 0.432699 * math.log( (155.0 * x) / 219.0 + 0.037584, 10.0) + 0.616596 ) + 0.03
		
	def sLog3(x) :
		'''
		Not yet implemented. See http://community.sony.com/sony/attachments/sony/large-sensor-camera-F5-F55/12359/2/TechnicalSummary_for_S-Gamut3Cine_S-Gamut3_S-Log3_V1_00.pdf .
		'''
		return x
		
	def DanLog(x) :
		return (10.0 ** ((x - 0.385537) / 0.2471896) - 0.071272) / 3.555556 if x > 0.1496582 else (x - 0.092809) / 5.367655
